#!/usr/bin/env python
"""
CommonMultiProcess

Common library

MultiProcess base structure
"""
from ductape.common.commonthread import CommonThread
from multiprocessing.queues import Queue
import logging
import multiprocessing
import sys
if sys.version_info[0] < 3:
    def t():
        return None
    multiprocessing.get_context = t
import time

# Consumer borrowed from http://broadcast.oreilly.com/
# EINTR fix borrowed from Boyd Waters

logger = logging.getLogger('ductape.commonmultiprocess')

class SafeSleep(object):
    '''
    IOError safe sleep
    '''
    def sleep(self,seconds):
        '''
        Sleeps for a certain amount of seconds
        Raises an exception if too many errors are encountered
        '''
        dt = 1e-3
        while dt < 1:
            try:
                time.sleep(seconds)
                return
            except IOError:
                logger.warning('IOError encountered in SafeSleep sleep()')
                try:
                    time.sleep(dt)
                except:pass
                dt *= 2
                
        e = IOError('Unrecoverable error')
        raise e

class SafeQueue(Queue):
    '''
    IOError safe multiprocessing Queue
    '''
    def __init__(self):
        if sys.version_info[0] < 3:
            Queue.__init__(self)
        else:
            Queue.__init__(self, ctx=multiprocessing.get_context())
        
    def empty(self):
        '''
        Returns True if the Queue is empty, False otherwise
        Raises an exception if too many errors are encountered 
        '''
        dt = 1e-3
        while dt < 1:
            try:
                isEmpty = Queue.empty(self)
                return isEmpty
            except IOError:
                logger.warning('IOError encountered in SafeQueue empty()')
                try:
                    time.sleep(dt)
                except:pass
                dt *= 2
                
        e = IOError('Unrecoverable error')
        raise e

    def get(self):
        '''
        Get the element in the queue
        Raises an exception if it's empty or if too many errors are
        encountered
        '''
        dt = 1e-3
        while dt < 1:
            try:
                element = Queue.get(self)
                return element
            except IOError:
                logger.warning('IOError encountered in SafeQueue get()')
                try:
                    time.sleep(dt)
                except:pass
                dt *= 2
                
        e = IOError('Unrecoverable error')
        raise e
    
    def put(self,element):
        '''
        Put the element in the queue
        Raises an exception if too many errors are
        encountered
        '''
        dt = 1e-3
        while dt < 1:
            try:
                Queue.put(self,element)
                return
            except IOError:
                logger.warning('IOError encountered in SafeQueue put()')
                try:
                    time.sleep(dt)
                except:pass
                dt *= 2
                
        e = IOError('Unrecoverable error')
        raise e

class Consumer(multiprocessing.Process):
    
    def __init__(self, 
                 task_queue = multiprocessing.Queue(),
                 result_queue = multiprocessing.Queue()):
        multiprocessing.Process.__init__(self)
        self.task_queue = task_queue
        self.result_queue = result_queue
        self.sleeper = SafeSleep()

    def run(self):
        while True:
            next_task = self.task_queue.get()
            self.sleeper.sleep(0.01)
            if next_task is None:
                # Poison pill means we should exit
                break
            answer = next_task()
            self.result_queue.put(answer)
        return
    
class CommonMultiProcess(CommonThread):
    '''
    Class CommonMultiProcess
    A Thread that can perform multiprocesses
    '''
    def __init__(self, ncpus=1, queue=None):
        if queue is None:
            if sys.version_info[0] < 3:
                queue = Queue()
            else:
                queue = Queue(ctx=multiprocessing.get_context())
        CommonThread.__init__(self, queue)
        
        self.ncpus = int(ncpus)
        # Parallelization
        self._parallel = None
        self._paralleltasks = SafeQueue()
        self._parallelresults = SafeQueue()
        self.sleeper = SafeSleep()
        
        # ID
        self._unique = 0
        
    def getUniqueID(self):
        self._unique += 1
        return self._unique
    
    def initiateParallel(self):
        self._parallel = [Consumer(self._paralleltasks,self._parallelresults)
                          for x in range(self.ncpus)]
        for consumer in self._parallel:
            consumer.start()
            
    def addPoison(self):
        for consumer in self._parallel:
            self._paralleltasks.put(None)

    def isTerminated(self):
        for consumer in self._parallel:
            if consumer.is_alive():
                return False
        return True

    def killParallel(self):
        for consumer in self._parallel:
            consumer.terminate()



#!/usr/bin/env python
"""
CommonThread

Common library

Thread base structure
"""

import sys
if sys.version_info[0] < 3:
    import Queue as queue
else:
    import queue
import logging
import os
import shutil
import threading
import time

__author__ = "Marco Galardini"

################################################################################
# Log setup

logger = logging.getLogger('ductape.commonthread')

################################################################################
# Classes

class Status(object):
    '''
    Class Status
    Gives informations about the run status of a specific thread
    '''
    def __init__(self,status=None,msg=None,maxstatus=None,
                    substatus=None,submsg=None,maxsubstatus=None,
                    fail=False):
        self.status = status
        self.msg = msg
        self.maxstatus = maxstatus
        #
        self.substatus = substatus
        self.submsg = submsg
        self.maxsubstatus = maxsubstatus
        # Fail msg?
        self.fail = fail

class CommonThread(threading.Thread):
    '''
    Class CommonThread: Common operations for a threading class
    '''
    _statusDesc = {0:'Not started',
               1:'Making room', 
               3:'Cleaning up'}
    
    _substatuses = []
    
    def __init__(self,queue=queue.Queue()):
        threading.Thread.__init__(self)
        # Thread
        self.msg = queue
        self._status = 0
        self._maxstatus = len(self._statusDesc)
        self._substatus = 0
        self._maxsubstatus = 0
        self._room = None
        self.killed = False
        
    def getStatus(self):
        return self._statusDesc[self._status]
    
    def getMaxStatus(self):
        return self._maxstatus
    
    def getMaxSubStatus(self):
        return self._maxsubstatus
    
    def getSubStatuses(self):
        return self._substatuses
    
    def resetSubStatus(self):
        self._substatus = 0
        self._maxsubstatus = 0
        
    def makeRoom(self,location=''):
        '''
        Creates a tmp directory in the desired location
        '''
        try:
            path = os.path.abspath(location)
            path = os.path.join(path, 'tmp')
            self._room = path
            os.mkdir(path)
        except:
            logger.debug('Temporary directory creation failed! %s'
                          %path)
    
    def startCleanUp(self):
        '''
        Removes the temporary directory
        '''
        if os.path.exists(self._room):
            logger.debug('Removing the old results directory (%s)'%
                         self._room)
            shutil.rmtree(self._room, True)
    
    def cleanUp(self):
        '''
        Removes the temporary directory
        '''
        shutil.rmtree(self._room, True)
        
    def run(self):
        self.updateStatus()
        self.makeRoom()

        self.updateStatus()
        self.cleanUp()
            
    def sendFailure(self,detail='Error!'):
        msg = Status(fail=True,
                     msg=detail)
        self.msg.put(msg)
        # Give some time for the message to arrive
        time.sleep(0.1)
        
    def updateStatus(self,sub=False,send=True):
        if not sub:
            self._status += 1
        if not send:
            return
        if self._status in self._substatuses:
            msg = Status(status=self._status,msg=self.getStatus(),
                         maxstatus=self.getMaxStatus(),
                         substatus=self._substatus,
                         maxsubstatus=self.getMaxSubStatus())
        else:
            msg = Status(status=self._status,msg=self.getStatus(),
                         maxstatus=self.getMaxStatus())
        self.msg.put(msg)
        
    def kill(self):
        self.killed = True



#!/usr/bin/env python
"""
Blast

Genome library

Classes to handle Blast analysis against a local database
"""
import logging
import os
import subprocess
import sys
if sys.version_info[0] < 3:
    from StringIO import StringIO # Python 2
else:
    from io import StringIO # Python 3

__author__ = "Marco Galardini"

################################################################################
# Log setup

logger = logging.getLogger('ductape.blast')

################################################################################
# Classes

# Useful class for parsing
class BlastHit:
    def __init__(self,query,align,hsp):
        '''
        Query, Alignment and Hsp are all Biopython objects derived from
        Blast results parsing
        '''
        self.query = query.query
        self.query_id = query.query.split(' ')[0]
        self.query_len = int(query.query_length)
        self.hit = align.hit_id
        self.hit_desc = align.hit_def
        self.hit_len = int(align.length)
        self.identity = float(hsp.identities) / float(hsp.align_length)
        self.align_len = int(hsp.align_length)
        self.mismatches = int(hsp.align_length - hsp.identities - hsp.gaps)
        self.gaps = int(hsp.gaps)
        self.query_start = int(hsp.query_start)
        self.query_end = int(hsp.query_end)
        self.subjct_start = int(hsp.sbjct_start)
        self.subjct_end = int(hsp.sbjct_end)
        self.evalue = float(hsp.expect)
        self.bits = float(hsp.bits)
        
    def getHomologyIndex(self):
        '''
        Get an Index useful for stating the quality of the homology measure
        '''
        import math
        HI=( (math.pow(self.identity,2)*(float(self.hit_len)) /
            (float(self.query_len))*(float(self.align_len)/float(self.query_len)))
            )
        return HI
    
    def getHitCoverage(self):
        '''
        Get the hit coverage
        '''
        return float(float(self.align_len)/float(self.hit_len))
    
    def getQueryCoverage(self):
        '''
        Get the query coverage
        '''
        return float(float(self.align_len)/float(self.query_len))
    
    def getKO(self):
        '''
        Assuming that this hit derives from a KEGG DB
        Returns the KO ID
        '''
        import re
        a=re.search("K[0-9]{1,}",
                    self.hit_desc)
        if a is not None:
            return a.group()
        else:
            return None
        
class Blaster(object):
    def __init__(self, useDisk=False):
        self._hits = None
        self._out = ''
        
        # No-disk
        self._useDisk = bool(useDisk)
        self.retrieved = ''
        self.query = ''
        self.out = ''
        
    def createDB(self,seqFile,dbType,outFile='BlastDB',parseIDs=True,
                        title='Generic Blast DB'):
        '''Generation of a Blast DB'''
        cmd = ('makeblastdb -in %s -dbtype %s -out %s -title "%s"')
        cmd = cmd%(seqFile,dbType,outFile,title)
        if parseIDs:
            cmd = cmd+' -parse_seqids'
        logger.debug('Create Blast DB cmd: %s'%cmd)
        proc = subprocess.Popen(cmd,shell=(sys.platform!="win32"),
                    stdin=subprocess.PIPE,stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE)
        out = proc.communicate()
        return_code = proc.returncode
        if return_code != 0:
            logger.warning('Blast DB creation failed with error %d'
                            %return_code)
            logger.warning('%s'%str(out[1]))

        return bool(not return_code)
    
    def retrieveFromDB(self, db, accession, out='out.fsa', isFile=False):
        '''Retrieve the desired sequence(s) from a Blast DB'''
        if not isFile:
            cmd=('blastdbcmd -db %s -entry "%s" -long_seqids'
                 %(db,accession))
        else:
            cmd=('blastdbcmd -db %s -entry_batch "%s" -long_seqids'
                 %(db,accession))
        
        if self._useDisk:
            cmd += ' > %s'%out
        
        logger.debug('BlastDBcmd cmd: %s'%cmd)
        proc = subprocess.Popen(cmd,shell=(sys.platform!="win32"),
                    stdin=subprocess.PIPE,stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE)
        out = proc.communicate()
        
        if not self._useDisk:
            self.retrieved = out[0]
        
        return_code = proc.returncode
        if return_code != 0:
            logger.warning('BlastDBcmd failed with error %d'
                            %return_code)
            logger.warning('%s'%str(out[1]))

        return bool(not return_code)
    
    def runBlast(self, queryFile, db, outFile='', evalue = 10,
                    task = '', ncpus = 1, additional = '', outfmt='5'):
        '''Run Blast with the desired parameters'''
        # Create the command line
        from Bio.Blast.Applications import NcbiblastpCommandline
        self._out = outFile
        cmd = NcbiblastpCommandline(db=db,
                evalue=float(evalue),
                outfmt=outfmt,
                num_threads=ncpus)
        if self._useDisk:
            cmd.set_parameter('query', queryFile)
            if outFile != '':
                cmd.set_parameter('out', outFile)
        if task != '':
            cmd.set_parameter('task', task)
        if additional !='':
            cmd = str(cmd)+' '+additional
        cmd=str(cmd)
        logger.debug('Run Blast cmd: %s'%cmd)
        # Run Blast and check the return code
        proc = subprocess.Popen(cmd,shell=(sys.platform!="win32"),
                    stdin=subprocess.PIPE,stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE)
        
        if not self._useDisk:
            if isinstance(self.query, str):
                proc.stdin.write(self.query.encode())
            else:
                proc.stdin.write(self.query)
                    
        out = proc.communicate()
        
        if not self._useDisk:
            self.out = out[0]
        
        return_code = proc.returncode
        if return_code != 0:
            logger.warning('Run Blast failed with error %d'
                            %return_code)
            logger.warning('%s'%str(out[1]))

        return bool(not return_code)
    
    def parseBlast(self, fileOut):
        '''Parse the xml blast output -- default file is self._out'''
        from Bio.Blast import NCBIXML

        if self._useDisk:
            self._out = fileOut
            handle = open(fileOut)
        else:
            handle = StringIO(self.out.decode('utf-8'))
            
        self._hits = NCBIXML.parse(handle)
        
    def getHits(self,expect=10.0):
        '''Returns a Generator query -> BlastObj'''
        if self._hits == None:
            self.parseBlast(self._out)
        for BlastQuery in self._hits:
            hits = []
            for alignment in BlastQuery.alignments:
                for hsp in alignment.hsps:
                    if float(hsp.expect) > expect:continue
                    # Save the hit details
                    h=BlastHit(BlastQuery,alignment,hsp)
                    hits.append(h)
            yield hits
            
class RunBBH(object):
    def __init__(self, query, queryid,
                 source, target, targetorg,
                 evalue, matrix, short = False, uniqueid = 1,
                 kegg = False, ko_entry = None, ko_id = None, useDisk=True):
        for x in [query, queryid, source, target, targetorg,
                  evalue, matrix, short, uniqueid,
                  kegg, ko_entry, ko_id, useDisk]:
            if isinstance(x, bytes):
                x = x.decode('utf-8')
        self.query = query
        self.queryid = queryid
        self.source = source
        self.target = target
        self.targetorg = targetorg
        self.evalue = evalue
        self.matrix = matrix
        self.short = short
        self.uniqueid = uniqueid
        self.kegg = kegg
        self.ko_entry = ko_entry
        self.ko_id = ko_id
        self.useDisk = bool(useDisk)
        
        self.out = self.queryid + '_' + str(self.uniqueid) +'.xml'
        self.blaster = Blaster(useDisk=self.useDisk)
        self.additional = (' -soft_masking true -dbsize 500000000 '+
                    '-use_sw_tback -max_target_seqs 1 -matrix %s'%self.matrix)
        
        if not self.useDisk:
            self.blaster.query = self.query
            self.queryreturn = ''
        else:
            self.queryreturn = self.query + '_' + str(self.uniqueid) + '_return'
    
    def _firstRun(self):
        if self.short:
            res = self.blaster.runBlast(self.query, self.target, self.out,
                         evalue = self.evalue,
                         task='blastp-short',
                         additional=self.additional)
        else:
            res = self.blaster.runBlast(self.query, self.target, self.out,
                         evalue = self.evalue,
                         additional = self.additional)
        
        return res
    
    def _secondRun(self, hit_len = None):
        # Second Blast run
        if not hit_len:
            if self.short:
                hit_len = 29
            else:
                hit_len = 100
                
        if not self.useDisk:
            self.blaster.query = self.blaster.retrieved
                
        if hit_len < 30:
            res = self.blaster.runBlast(self.queryreturn, self.source, self.out,
                     evalue = self.evalue,
                     task='blastp-short',
                     additional=self.additional)
        else:
            res = self.blaster.runBlast(self.queryreturn, self.source, self.out,
                     evalue = self.evalue,
                     additional=self.additional)
            
        return res
    
    def __call__(self):
        if not self.kegg:
            # First Blast run
            res = self._firstRun()
            
            if not res:
                if self.useDisk:
                    try:
                        os.remove(self.out)
                    except:pass
                    
                return [None, self.targetorg, False]
            
            self.blaster.parseBlast(self.out)
            for hits in self.blaster.getHits(self.evalue):
                if len(hits) == 0:
                    break
                targethit = hits[0]
    
                if not self.blaster.retrieveFromDB(self.target, targethit.hit,
                                      out=self.queryreturn):
    
                    if self.useDisk:
                        try:
                            os.remove(self.out)
                        except:pass
                        
                    return [None, self.targetorg, False]
    
                # Second Blast run            
                res = self._secondRun(targethit.hit_len)
                break
        else:
            if not self.blaster.retrieveFromDB(self.target, self.ko_entry,
                                      out=self.queryreturn):
                if self.useDisk:
                    try:
                        os.remove(self.out)
                    except:
                        pass
                return [None, self.targetorg, False]
            res = self._secondRun()
        
        if not res:
        
            if self.useDisk:
                try:
                    os.remove(self.out)
                    os.remove(self.queryreturn)
                except:pass
            return [None, self.targetorg, False]
        
        self.blaster.parseBlast(self.out)
        for hits in self.blaster.getHits(self.evalue):
            if len(hits) == 0:
                return [None, self.targetorg, True]
            sourcehit = hits[0]
            if self.queryid == sourcehit.hit:
                if self.useDisk:
                    os.remove(self.out)
                    os.remove(self.queryreturn)
                if self.kegg:
                    return [self.ko_id,self.queryid, True]
                else:
                    return [sourcehit.query_id.replace('lcl|',''),
                        self.targetorg, True]
            else:
                if self.useDisk:
                    os.remove(self.out)
                    os.remove(self.queryreturn)
                return [None, self.targetorg, True]

        if self.useDisk:
            os.remove(self.out)
        return [None, self.targetorg, True]



#!/usr/bin/env python
"""
utils

Common library

Spare parts
"""
import logging
import time
import sys

################################################################################
# Log setup

logger = logging.getLogger('ductape.utils')

################################################################################
# Methods

# Borrowed from: www.garyrobinson.net
def slice_it(li, cols=10):
    start = 0
    for i in range(cols):
        stop = start + len(li[i::cols])
        yield li[start:stop]
        start = stop
        
def get_span(li, span=4):
    i = 0
    while i < len(li):
        yield li[i : i+span]
        i += span

# Borrowed from stackoverflow
# python-most-idiomatic-way-to-convert-none-to-empty-string
def xstr(s):
    if s is None:
        return ''
    return str(s)

# Borrowed from http://www.scipy.org/Cookbook/SignalSmooth
def smooth(x,window_len=11,window='hanning'):
    """smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    input:
        x: the input signal
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal

    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)

    see also:

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter

    TODO: the window parameter could be the window itself if an array instead of a string
    """
    from numpy.core.numeric import ones
    import numpy
    
    x = numpy.array(x)

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
    #if len(x) < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")


    if window_len<3:
        return x


    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")


    s=numpy.r_[2*x[0]-x[window_len:1:-1],x,2*x[-1]-x[-1:-window_len:-1]]
    #print(len(s))
    if window == 'flat': #moving average
        w=ones(window_len,'d')
    else:
        w=eval('numpy.'+window+'(window_len)')

    y=numpy.convolve(w/w.sum(),s,mode='same')
    return y[window_len-1:-window_len+1]

def compress(x, span=10):
    return [i[0] for i in get_span(x, span)]

def rgb_to_hex(rgb):
    return '#%02x%02x%02x' % rgb

def rangeColors(minimum, maximum, colorrange):
    '''
    Given a min and a max value, returns a dict with stepwise colors
    '''
    hexs = {}
    prev = '#FFFFFF'
    i = minimum
    for color in slice_it(colorrange, cols=int(maximum-minimum+1)):
        if len(color) == 0:
            hexs[i] = prev
        else:
            if i == minimum:
                hexs[i] = rgb_to_hex(tuple([int(round(x*255))
                              for x in color[0][:3]])).upper()
            else:
                hexs[i] = rgb_to_hex(tuple([int(round(x*255))
                              for x in color[-1][:3]])).upper()
        prev = hexs[i]
        i += 1
    
    return hexs

def makeRoom(location='', *args):
    '''
    Creates a tmp directory in the desired location
    Any extra argument is a nested directory
    '''
    import os
    path = os.path.abspath(location)
    path = os.path.join(path, 'tmp')
    try:os.mkdir(path)
    except:pass
    
    for name in args:
        path = os.path.join(path, name)
        try:os.mkdir(path)
        except:pass
    
    return path

def isOnline(url='http://8.8.8.8', timeout=1, retries=2):
    '''
    Check if a particular site is online (with timeout)
    If an IP address is provided, DNS lookup is skipped, so the timeout will be
    onoured.
    
    Thanks to @unutbu on Stack Overflow
    http://stackoverflow.com/questions/3764291/checking-network-connection
    '''
    attempts = 0
    while True:
        try:
            if sys.version_info[0] < 3:
                from urllib2 import urlopen
            else:
                from urllib.request import urlopen
            response=urlopen(url, timeout=1)
            return
        except Exception as e:
            attempts += 1
            logger.debug('test failed! Attempt %d'
                          %attempts)
            logger.debug('%s'%str(e))
            time.sleep(0.5*attempts)
            try:
                logger.debug(url)
            except:pass
            if attempts >= retries:
                logger.warning('Connection test failed!')
                raise IOError('Connection test failed!')

def safeSubtraction(a, b):
    '''
    Tries to perform a subtraction: if it doesn't work returns None
    '''
    try:
        return a - b
    except:
        return None



#!/usr/bin/env python
"""
Actions for terminal

DuctApe Library

All the actions required for the analysis (terminal utilitites)
"""
from ductape.storage.SQLite.database import Project, Kegg
import logging

__author__ = "Marco Galardini"

################################################################################
# Log setup

logger = logging.getLogger('ductape.actionsterm')

################################################################################
# Methods

def fetchKegg(project, keeptrying=False):
    from ductape.kegg.kegg import KeggNet, KeggAPI, BaseKegg
    from ductape.terminal import RunThread
    
    # Check if we have to fetch the whole kegg DB
    fetch = False
    proj = Project(project)
    
    logger.info('Checking connectivity')
    bk = BaseKegg()
    try:
        bk.checkConnection()
    except Exception as e:
        logger.error(str(e))
        return False
    
    k = KeggAPI()
    try:
        k.getDBVersion()
        release = k.result[1]
    except Exception as e:
        logger.warning('Could not fetch the KEGG DB version (%s)'%str(e))
        release = None
    if proj.isKegg():
        if release and proj.kegg < release:
            logger.warning('A new KEGG DB version is available (%s, was %s)'%
                           (str(release), str(proj.kegg)))
            fetch = True
    else:
        fetch = True
  
    if fetch:
        logger.info('Fetching the whole KEGG metabolic map')
        if release:
            logger.info('KEGG DB release %s'%str(release))
        kegg = Kegg(project)
        
        knet = KeggNet(keeptrying=keeptrying)
        if not RunThread(knet):
            return False
        
        # Details
        kegg.addPathways(knet.result.path)
        logger.info('Added %d Path IDs'%len(knet.result.path))
        kegg.addReactions(knet.result.react)
        logger.info('Added %d Re IDs'%len(knet.result.react))
        kegg.addCompounds(knet.result.comp)
        logger.info('Added %d Co IDs'%len(knet.result.comp))
        kegg.addRPairs(knet.result.rpair)
        logger.info('Added %d RPair IDs'%len(knet.result.rpair))
        # Links
        kegg.addPathReacts(knet.result.pathreact)
        kegg.addReactComps(knet.result.reactcomp)
        kegg.addCompReacts(knet.result.compreact)
        kegg.addPathComps(knet.result.pathcomp)
        kegg.addReactRPairs(knet.result.reactrpair)
        kegg.addRPairReacts(knet.result.rpairreact)
        logger.info('Added Kegg links')
        # HTML maps
        kegg.addPathHtml(knet.result.pathmaps)
        logger.info('Added Kegg maps')
        
        # Add the release version
        if release:
            proj.setKegg(release)
    else:
        logger.info('KEGG db is up-to-date')
        
    return True
